/*
 * batcher - send a bunch of news articles as an unbatch script
 *
 * Usage: batcher OUTFILE listfile <size>
 *
 *	where listfile is a file containing a list, one per line, of
 *	names of files containing articles.  The first field of each line
 * is looked at, and if present the second for filesize. If the second
 * field is not present, the file-size is computed by a lseek().
 * Non-absolute pathnames are understood to lie under the current
 * directory; chdiring to the right place is the parent's problem.
 */

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

char Buf[1024];
long __regargs process(register char *article, register FILE *fout);
void main(long argc, char *argv[]);

void main(long argc,char *argv[])
{	char outfile[256];
	char *listfile;
	int size, totalsize = 0;
	FILE *fin, *fout;

	freopen("NewsCtl:batch.errlog","a",stderr);

	if(argc != 4)
	{	fprintf(stderr,"USAGE: %s OUTFILEname LISTFILEname <digits> !\n",argv[0]);
		exit(20);
	}
	stccpy(outfile,argv[1],sizeof(outfile));	/* this program is intended to */
	listfile = argv[2];	/* be launched from another program, so we can */
	size = atol(argv[3]);	/* assume, that the parameters are correct ! */
	if(fout = fopen(outfile,"w"))
	{	if(fin = fopen(listfile,"r"))
		{	while(!((totalsize > size) || feof(fin) || ferror(fin) || ferror(fout)))
			{	fgets(Buf,sizeof(Buf),fin);
				if(Buf[0])
					totalsize += process(Buf,fout);
			}
			fclose(fout);
			if(feof(fin))
			{	fclose(fin);
				unlink(listfile);
			}
			else
			{	stccpy(outfile,listfile,sizeof(outfile)-4);
				strcat(outfile,".TMP");
				if(fout = fopen(outfile,"w"))
				{	while(!(feof(fin) || ferror(fin))) /* Keep remaining jobs... */
					{	fgets(Buf,sizeof(Buf),fin);
						fputs(Buf,fout);
					}
					fclose(fout);
					fclose(fin);
					unlink(listfile);
					rename(outfile,listfile);
				}
				else
				{	fprintf(stderr,"FATAL ERROR: can't update %s !!!\n",listfile);
					printf(			"FATAL ERROR: can't update %s !!!\n",listfile);
				}
			}
		}
		else
			fprintf(stderr,"can't open %s for reading !\n",listfile);
	}
	else
		fprintf(stderr,"can't open %s for writing !\n",outfile);
}

/*
 - process - process an article
 */
long __regargs process(register char *article, register FILE *fout)
{	register int artfile;
	register int count;
	long size;
	register char *endp;

	endp = strchr(article, '\t');
	if (endp == NULL)
		endp = strchr(article, ' ');
	if (endp == NULL)
		endp = strchr(article, '\n');
	if (endp == NULL)
		endp = article + strlen(article) - 1;
	else
		*endp = '\0';

	artfile = open(article, 0);
	if (artfile < 0)
   {	/*
		 * Can't read the article.  This isn't necessarily a
		 * disaster, since things like cancellations will do
		 * this.  Mumble and carry on.
		 */
		return(0);
	}

	size = atol(stpblk(endp+1));
   if(! size) /* If 'f'-TransmissionFlag is used*/
	{		/* , size contains the actual article-size, else 0 */
      size = lseek(artfile,0,2);
      lseek(artfile,0,0);
	}
   if(! size)
      return(0);
	fprintf(fout,"#! rnews %ld\n",size);
	while ((count = read(artfile, Buf, sizeof(Buf))) > 0)
		if (fwrite(Buf, sizeof(char), count,fout) != count)
			fprintf(stderr,"write failure in `%s'", article);
	if (count < 0)
		fprintf(stderr,"read failure in `%s'", article);

	(void) close(artfile);
	return(size);
}
